
"""
Base classes for the different methods in nystrompca.algorithms

"""

from .kernel         import Kernel
from .transformation import Transformation
from .regression     import Regression
from .kernel_machine import KernelMachine
from .nystrom_method import NystromMethod
